/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ReadVorbis.cpp
 */

#include "ReadVorbis.h"
#include <m_pd.h>

ReadVorbis::ReadVorbis()
{
  fp = NULL;
}


bool ReadVorbis::Initialize(const char *filename) 
{  
  vorbis_info *vi;
  
  if (fp != NULL) {
    fclose(fp); 
  }
  fp = fopen( filename,  "r"); 
  if (fp == NULL) {
    //fprintf(stderr,  "couldnt open file\n");
    return false;
  } 


  if(ov_open(fp, &vf, NULL, 0) < 0) {
    fprintf(stderr," does not appear to be an Ogg bitstream.\n");
    return false;
  }

  vi=ov_info(&vf,-1);
  samplerate = vi->rate;
  num_channels = vi->channels;
  //fprintf(stderr, "ReadVorbis: opening file: [%s] %d (Hz), %d chan(s)\n", 
  //  filename, vi->rate, vi->channels);
  return true;
}

int ReadVorbis::Decode(float *buffer, int size) {
  long ret;
  int current_section;
  float **buftmp;
  int x=0;
  if (CHUNKSIZE > size) return 0;
  ret=ov_read_float(&vf, &buftmp , CHUNKSIZE,  &current_section);

  if (num_channels == 1) {
    for(int j = 0; j < ret; j++)
      {
	buffer[x++] = buftmp[0][j];
      }
  } else {
    for(int j = 0; j < ret; j++)
      {
	buffer[x++] = buftmp[0][j];
	buffer[x++] = buftmp[1][j];
      }
  }
  return x;  //ret;

}


bool ReadVorbis::Rewind() {
  if ( ov_pcm_seek(&vf, 0) == 0)
    return true;
  else
    return false;
}

bool ReadVorbis::PCM_seek(long bytes) {
  if ( ov_pcm_seek(&vf, bytes) == 0)
    return true;
  else
    return false;

}

bool ReadVorbis::TIME_seek(double seconds) {
  if ( ov_time_seek(&vf, seconds) == 0)
    return true;
  else
    return false;
}
