/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * main.cpp
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <alloca.h>

#include "main.h"



readanysf::readanysf()
{ 
  fifo = new Fifo( FIFOSIZE  );
  requestcode = R_NOTHING;
  state = STATE_IDLE;
  loop = false;
  eof = true;
  format = -1;
  readsf = NULL;
  //stopfill = false;

  //SRC_SINC_BEST_QUALITY 
  //SRC_SINC_MEDIUM_QUALITY
  //SRC_SINC_FASTEST
  //SRC_ZERO_ORDER_HOLD
  //SRC_LINEAR
  src_mode = SRC_SINC_FASTEST;
  src_state=  NULL;
  src_data.input_frames =0;

  AddInAnything();  // add one inlet for any message
  AddOutSignal("audio out Left");
  AddOutSignal("audio out Right");
  AddOutBang();  // add one float outlet (has index 2)

  FLEXT_ADDMETHOD_(0,"start",m_start);
  FLEXT_ADDMETHOD_(0,"play",m_start);
  FLEXT_ADDMETHOD_(0,"pause",m_pause);
  FLEXT_ADDMETHOD_(0,"stop",m_stop);
  FLEXT_ADDMETHOD_(0,"open",m_open);
  FLEXT_ADDMETHOD_(0,"loop",m_loop_f);
  FLEXT_ADDMETHOD_(0,"pcm_seek",m_pcm_seek);
  FLEXT_ADDMETHOD_(0,"time_seek",m_time_seek);

  FLEXT_ADDMETHOD_(0,"speed",m_src_factor);
  FLEXT_ADDBANG(0,m_bang);
  FLEXT_CALLMETHOD(m_child);

  
  src_buffer = (float *)malloc ( (Blocksize())*2*FLOATSIZE );

  //post("Blocksize = %d", Blocksize());
} 


readanysf::~readanysf()
{

  mutex.Lock();
  quit = true;
  cond.Signal();
  state = STATE_IDLE;
  if (readsf != NULL)
    delete(readsf);
  if (fifo != NULL)
    delete(fifo);

  mutex.Unlock();

  StopThreads();
}

void readanysf::m_bang() {
  ToOutBang(2);
}


void readanysf::m_loop_f(float f) {
  if (f == 0.0) {
    post("looping Off");
    mutex.Lock();
    loop = false;
    mutex.Unlock();
    return;
  } else {
    post("looping On");
    mutex.Lock();
    loop = true;
    mutex.Unlock();
    return;
  }
}


void readanysf::m_start() {
  mutex.Lock();
  if (state != STATE_STREAM) {
    if (filename[0] == 0 ) {
      post("first open a file");
      mutex.Unlock();
      return;
    }
    if ( requestcode != R_OPEN ) {

      eof = false;
      bzero((void *)src_buffer, sizeof(src_buffer));
      state = STATE_STREAM;
      //stopfill = false;
      requestcode = R_PROCESS;
    } else {
      post("still initializing...");
    }
  } else {
    post("already playing");
  }
  mutex.Unlock();
  cond.Signal();
  return;
}

void readanysf::m_time_seek(float f) {
  mutex.Lock();
  if (f > 0.0) 
    { timeseek = (double)f; requestcode = R_PROCESS; }
  mutex.Unlock();
  cond.Signal();
}

void readanysf::m_pcm_seek(int i) {
  mutex.Lock();
  if (i > 0)
    { pcmseek = (long) i; requestcode = R_PROCESS;}
  mutex.Unlock();
  cond.Signal();
}

void readanysf::m_stop() {
  mutex.Lock();
  //stopfill = true;
  requestcode = R_STOP;
  state = STATE_IDLE;
  mutex.Unlock();
  cond.Signal();
}

void readanysf::m_pause() {
  mutex.Lock();
  state = STATE_IDLE;
  mutex.Unlock();
}

void readanysf::m_open(t_symbol *s) {
  cond.Signal();
  sprintf(filename,GetString(s));
  if (filename[0] == 0) {
    post("bogus filename?");
    return;
  } else {
    mutex.Lock();
    if ( requestcode != R_OPEN ) {
      post("opening: %s", filename);
      state = STATE_IDLE;
      requestcode = R_OPEN;
    } else {
      post("still initailizing old file, please be patient");
    }
    mutex.Unlock();
    cond.Signal();
  }
}

void readanysf::m_src_factor(float f) {
  if (src_is_valid_ratio (f)) {
    mutex.Lock();
    src_factor = (double)f ;
    mutex.Unlock();
  }
}


void readanysf::FillFifo() {
  int ret, wret;

  //mutex is locked already
  if (pcmseek) {
    if (readsf !=NULL) {
      if ( readsf->PCM_seek(pcmseek) ) {
	fifo->Flush(); requestcode = R_PROCESS; pcmseek =0;
      }
    }
  }
  if (timeseek != 0.0){
    if (readsf !=NULL) {
      if ( readsf->TIME_seek( timeseek ) ) {  
	fifo->Flush(); requestcode = R_PROCESS; timeseek = 0.0;
       }
    }
  }
  
  while ( fifo->FreeSpace() > INOUTSIZE) {  // leave enough space for odd chunks
    mutex.Unlock();
    //post("decoding ...");
    ret =  readsf->Decode(read_buffer, READBUFFER);
    mutex.Lock();

    if (ret > 0) {
      wret = fifo->Write( (void *)read_buffer, ret*FLOATSIZE); //post("wrote %d", wret);
    } else {                         //post("eof, %d", ret);
      if ( readsf->Rewind() ) {      //post("Rewound the file");
        if (!loop) {
          //state = STATE_IDLE;      //this is premature. we check for eof in m_signal routine
          eof = true;
          requestcode = R_NOTHING;
          break;
        }
      }
    } 
  }
}


void readanysf::m_child()
{
  quit = false;

  while(!quit ) {   //while(!ShouldExit()) {

    mutex.Lock();
    if ( requestcode == R_STOP ) {
      fifo->Flush();
      readsf->Rewind();
      requestcode = R_PROCESS;
      state = STATE_IDLE;
      //post ("stop");
      FillFifo();
      mutex.Unlock();

    } else if ( requestcode == R_OPEN ) {
      //post("just inside R_OPEN");

      if (readsf != NULL) { 
	delete(readsf);
      }
      mutex.Unlock();
      format = SetType(filename);
      mutex.Lock();

      fifo->Flush();
      if ( format >= 0) {

	mutex.Unlock();
	if ( readsf->Initialize(filename) ) {
	  mutex.Lock();
	  num_channels = readsf->get_channels();
	  if (num_channels > 2) num_channels = 2;
	  samplerate = readsf->get_samplerate();
	  src_factor = Samplerate()/samplerate;
	  if ( !src_is_valid_ratio (src_factor) )
	    src_factor = 1.0;
	  requestcode = R_PROCESS;
	  mutex.Unlock();

	} else {
	  mutex.Lock();
	  requestcode = R_NOTHING;
	  if (readsf != NULL)
	    delete(readsf);
	  readsf = NULL;
	  filename[0] = 0;
	  mutex.Unlock();
	  m_bang(); 

	}
      } else { // not a recognized file type
	post("file not recognized");
	filename[0] = 0;
	state = STATE_IDLE;
	requestcode = R_NOTHING;
	mutex.Unlock();
	m_bang(); 

      }

    } else if ( requestcode ==  R_PROCESS ) {
      //post("waiting..process");
      FillFifo(); //take care of mutex locking in FillFifo routine
      mutex.Unlock();

      cond.Wait();
    
    } else { //NOTHING
      mutex.Unlock();
      post("waiting..nothing");
      cond.Wait();
    }
  }
  return;
}


int readanysf::m_resample(int frames) {
  unsigned int size, out_gen; 
 

  if (src_factor == 1.0) {
    size = frames * num_channels * FLOATSIZE ;
    fifo->Read( (void *)src_buffer, size);
    return size/FLOATSIZE;
  }

  if (src_channels != num_channels) {
    src_state = src_delete (src_state) ;
    src_state=  src_new (src_mode, num_channels, &src_error) ;
    src_channels = num_channels;
  }

  src_data.output_frames = frames ;  
  out_gen = 0;

  while (src_data.output_frames > 0) {
    if (src_data.input_frames <= 0) {
      if ( src_factor > 1.0 )
	size = frames * num_channels * FLOATSIZE ;
      else size = 1024 * num_channels * FLOATSIZE ;
      fifo->Read( (void *)pd_buffer, size);
      src_data.data_in = pd_buffer;
      src_data.data_out = src_buffer;
      src_data.input_frames =  size/FLOATSIZE/num_channels;
    }
    
    src_data.src_ratio = src_factor;
    if ( src_error = src_process ( src_state, &src_data) ) { 
      post("SRC error: %s", src_strerror (src_error));
      return 0;
    } else {
      if ( src_data.output_frames_gen == 0) {
	break;
      }
      src_data.data_in += src_data.input_frames_used*num_channels ;
      src_data.input_frames -= src_data.input_frames_used ;
      src_data.output_frames -= src_data.output_frames_gen;
      out_gen += src_data.output_frames_gen;
    }
  }
  //if (out_gen != frames) {
  //post("outgen %d, frames %d", out_gen, frames);
  //}

  return out_gen*num_channels;
}


FLEXT_NEW_DSP("readanysf~", readanysf)

void readanysf::m_signal(int n, float *const *in, float *const *out)
{
  float *outs1  = out[0];
  float *outs2  = out[1];
  
  mutex.Lock();

  if (state == STATE_STREAM) {

    int ch = num_channels;
    int size = m_resample(n);
    
    if (size < n*num_channels) {
      if (eof) { 
	post("eof");
	state = STATE_IDLE;
	fifo->Flush();
	requestcode = R_PROCESS;  
	m_bang();  //  bang out at end of file
      } else {
	//post("not EOF, but buffer is empty");
      }
      //if not eof, buffer is stuck somehow so lets keep crunching. 
      //in any case fill from size returned to the end with zeros
      for (int x = size; x < (n*num_channels); x+=num_channels){
	*outs1++  = 0.0;
	*outs2++  = 0.0;
      }
    }

    if ( fifo->FreeSpace() > INOUTSIZE ) {
      if ( requestcode == R_PROCESS && !eof) {
	//post("fifo free: %d", fifo->FreeSpace());
	cond.Signal();  // tell the child that the fifo is hungry
      }
    }
    mutex.Unlock(); 
    if (ch > 1) {
      for (int x = 0; x < size; x+=2) {
	*outs1++  = src_buffer[x];
	*outs2++  = src_buffer[x+1];
      }
    } else { //mono
      for (int x = 0; x < size; x++) {
	*outs1++  = src_buffer[x];
	*outs2++  = src_buffer[x];
      }
    }
    
  } else { // we're not streaming.  just zero out the audio outlets
    mutex.Unlock();	
    while (n--) {
      *outs1++  = 0.0;
      *outs2++  = 0.0;
    }
  }
}  


int readanysf::SetType(const char* thefile) {

  char buf[18];
  int fd, fmt;


  
  if( !strncasecmp(thefile ,"http://",7) ) {
    fmt = FORMAT_HTTP_MP3; 
    readsf = new ReadMadUrl();
    return fmt;
  }
  

  fd = open( thefile, O_RDONLY);
  if (fd == -1) {
    fmt = -1;
    readsf = NULL;
    close(fd);
    return -1;
  }

  int bytesread = read(fd, buf, 16);

  if (bytesread < 4) {
    readsf = NULL;
    close(fd);
    return -1;
  }


  if (!strncmp(buf, ".snd", 4)) {
    fmt = FORMAT_NEXT; //, bigendian = 1;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "dns.", 4)) {
    fmt = FORMAT_NEXT; //, bigendian = 0;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "RIFF", 4)) {
    if (bytesread < 12 || strncmp(buf + 8, "WAVE", 4)) {
      readsf = NULL;
      post("bad header ?");
      return -1;
    }
    fmt = FORMAT_WAVE; //, bigendian = 0;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "FORM", 4)) {
    if (bytesread < 12 || strncmp(buf + 8, "AIFF", 4))
      return -2; //goto badheader;
    fmt = FORMAT_AIFF; //, bigendian = 1;
    close(fd);
    readsf = new ReadRaw();
    return fmt;
  } else if (!strncmp(buf, "OggS", 4) ) {
    fmt = FORMAT_VORBIS;
    close(fd);
    readsf = new ReadVorbis();
    return fmt;
  } else if (!strncmp(buf, "ID3", 3)) {
    fmt = FORMAT_MAD;
    close(fd);
    readsf = new ReadMad();
    return fmt;
  } else if (!strncasecmp(buf, "FLAC", 4)) {
    // } else if( !strncasecmp(thefile+strlen(thefile)-4,".fla",4) ) {
    fmt = FORMAT_FLAC;  
    close(fd);
    readsf = new ReadFlac();
    return fmt;
  } else {
    unsigned int sync, bitrate;
    sync     = (unsigned char)buf[0];
    sync     = sync << 3;
    sync     |= ( (unsigned char)buf[1] & 0xE0) >> 5;
    if ( sync == 0x7FF ) {
      fmt = FORMAT_MAD; //printf("mp3 at the beginning\n");
      close(fd);
      readsf = new ReadMad();
      return fmt;
    } else if( !strncasecmp(thefile+strlen(thefile)-4,".mp3",4) ) {
      //trust that its mp3
      fmt = FORMAT_MAD;  //printf("*.mp3 file ending detected....trying mp3\n");
      close(fd);
      readsf = new ReadMad();
      return fmt;
    }
  }
  readsf = NULL;
  return -1;
    
}
